from lsprotocol.types import TEXT_DOCUMENT_COMPLETION
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionParams
from lsprotocol.types import ShowDocumentParams
from pygls.server import LanguageServer

server = LanguageServer("window-show-document", "v1")


@server.feature(TEXT_DOCUMENT_COMPLETION)
async def completion(ls: LanguageServer, params: CompletionParams):
    items = []
    await ls.show_document_async(ShowDocumentParams(uri=params.text_document.uri))

    for i in range(10):
        items.append(CompletionItem(label=f"item-{i}"))

    return items


if __name__ == "__main__":
    server.start_io()



import sys

from lsprotocol.types import ClientCapabilities
from lsprotocol.types import CompletionList
from lsprotocol.types import CompletionParams
from lsprotocol.types import InitializeParams
from lsprotocol.types import Position
from lsprotocol.types import TextDocumentIdentifier

import pytest
import pytest_lsp
from pytest_lsp import ClientServerConfig
from pytest_lsp import LanguageClient


@pytest_lsp.fixture(
    config=ClientServerConfig(server_command=[sys.executable, "server.py"]),
)
async def client(lsp_client: LanguageClient):
    # Setup
    params = InitializeParams(capabilities=ClientCapabilities())
    await lsp_client.initialize_session(params)

    yield

    # Teardown
    await lsp_client.shutdown_session()


@pytest.mark.asyncio
async def test_completions(client: LanguageClient):
    results = await client.text_document_completion_async(
        params=CompletionParams(
            position=Position(line=1, character=0),
            text_document=TextDocumentIdentifier(uri="file:///path/to/file.txt"),
        )
    )

    assert results is not None

    if isinstance(results, CompletionList):
        items = results.items
    else:
        items = results

    labels = [item.label for item in items]
    assert labels == [f"item-{i}" for i in range(10)]

    for idx, shown_message in enumerate(client.messages):
        assert shown_message.message == f"Suggesting item {idx}"



import inspect
import logging
import sys
import textwrap
import typing
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

import attrs
import pytest
import pytest_asyncio
from pygls.client import JsonRPCClient

from pytest_lsp.client import LanguageClient
from pytest_lsp.client import make_test_lsp_client

logger = logging.getLogger("client")


@attrs.define
class ClientServerConfig:
    """Configuration for a Client-Server connection."""

    server_command: List[str]
    """The command to use to start the language server."""

    client_factory: Callable[[], JsonRPCClient] = attrs.field(
        default=make_test_lsp_client,
    )
    """Factory function to use when constructing the test client instance."""

    server_env: Optional[Dict[str, str]] = attrs.field(default=None)
    """Environment variables to set when starting the server."""

    def _get_devtools_command(self, server: str) -> List[str]:
        """Get the lsp-devtools command required to connect to the given ``server``"""

        if ":" in server:
            host, port = server.split(":")
        else:
            host, port = "localhost", server

        try:
            int(port)
        except ValueError as e:
            raise ValueError(f"Invalid port number: {port!r}") from e

        return ["lsp-devtools", "agent", "--host", host, "--port", port, "--"]

    def get_server_command(self, devtools: Optional[str] = None) -> List[str]:
        """Get the command to start the server with."""
        server_command = []
        if devtools is not None:
            server_command.extend(self._get_devtools_command(devtools))

        server_command.extend(self.server_command)

        return server_command

    async def start(self, devtools: Optional[str] = None) -> JsonRPCClient:
        """Return the client instance to use for the test.

        Parameters
        ----------
        devtools
           If set, enable ``lsp-devtools`` integration, should be of the form
           ``<port>`` or ``<host>:<port>``. Where ``<host>`` and ``<port>``
           describe how to connect to an ``lsp-devtools`` server program.

        Returns
        -------
        JsonRPCClient
           The client instance to use in the test.
        """
        client = self.client_factory()
        server_command = self.get_server_command(devtools)

        await client.start_io(*server_command, env=self.server_env)
        return client


def pytest_addoption(parser):
    group = parser.getgroup("lsp")
    group.addoption(
        "--lsp-devtools",
        dest="devtools",
        nargs="?",
        action="store",
        default=None,
        const="localhost:8765",
        help="Enable lsp-devtools integration.",
    )


def pytest_runtest_makereport(item: pytest.Item, call: pytest.CallInfo):
    """Add any captured log messages to the report."""
    client: Optional[LanguageClient] = None

    if not hasattr(item, "funcargs"):
        return

    for arg in item.funcargs.values():
        if isinstance(arg, LanguageClient):
            client = arg
            break

    if not client:
        return

    levels = ["ERROR: ", " WARN: ", " INFO: ", "  LOG: "]

    if call.when == "setup":
        captured_messages = client.log_messages[: client._setup_log_index + 1]
    else:
        captured_messages = client.log_messages[client._last_log_index :]
        client._last_log_index = len(client.log_messages)

    messages = [
        f"{textwrap.indent(m.message, levels[m.type.value - 1])}"
        for m in captured_messages
    ]

    if len(messages) > 0:
        item.add_report_section(call.when, "window/logMessages", "\n".join(messages))


# anext() was added in 3.10
if sys.version_info.minor < 10:

    async def anext(it):
        return await it.__anext__()


def get_fixture_arguments(
    fn: Callable,
    client: JsonRPCClient,
    request: pytest.FixtureRequest,
) -> dict:
    """Return the arguments to pass to the user's fixture function.

    Parameters
    ----------
    fn
       The user's fixture function

    client
       The test client instance to inject

    request
       pytest's request fixture

    Returns
    -------
    dict
       The set of arguments to pass to the user's fixture function
    """
    kwargs: Dict[str, Any] = {}
    required_parameters = set(inspect.signature(fn).parameters.keys())

    # Inject the 'request' fixture if requested
    if "request" in required_parameters:
        kwargs["request"] = request
        required_parameters.remove("request")

    # Inject the language client
    for name, cls in typing.get_type_hints(fn).items():
        if issubclass(cls, JsonRPCClient):
            kwargs[name] = client
            required_parameters.remove(name)

    # Assume all remaining parameters are pytest fixtures
    for name in required_parameters:
        kwargs[name] = request.getfixturevalue(name)

    return kwargs


def fixture(
    fixture_function=None,
    *,
    config: ClientServerConfig,
    **kwargs,
):
    """Define a fixture that returns a client connected to a server running in a
    background sub-process

    Parameters
    ----------
    config
       Configuration for the client and server.
    """

    def wrapper(fn):
        @pytest_asyncio.fixture(**kwargs)
        async def the_fixture(request):
            devtools = request.config.getoption("devtools")
            client = await config.start(devtools=devtools)

            kwargs = get_fixture_arguments(fn, client, request)
            result = fn(**kwargs)
            if inspect.isasyncgen(result):
                try:
                    await anext(result)
                except StopAsyncIteration:
                    pass

            yield client

            if inspect.isasyncgen(result):
                try:
                    await anext(result)
                except StopAsyncIteration:
                    pass

            await client.stop()

        return the_fixture

    if fixture_function:
        return wrapper(fixture_function)

    return wrapper



from typing import Any

import pytest
from lsprotocol import types

from pytest_lsp import checks

a_range = types.Range(
    start=types.Position(line=1, character=0),
    end=types.Position(line=2, character=0),
)


@pytest.mark.parametrize(
    "capabilities,method,params,expected",
    [
        (
            types.ClientCapabilities(
                window=types.WindowClientCapabilities(work_done_progress=False)
            ),
            types.WINDOW_WORK_DONE_PROGRESS_CREATE,
            types.WorkDoneProgressCreateParams(token="id-123"),
            "does not support 'window/workDoneProgress/create'",
        ),
        (
            types.ClientCapabilities(
                workspace=types.WorkspaceClientCapabilities(configuration=False)
            ),
            types.WORKSPACE_CONFIGURATION,
            types.WorkspaceConfigurationParams(items=[]),
            "does not support 'workspace/configuration'",
        ),
    ],
)
def test_params_check_warning(
    capabilities: types.ClientCapabilities, method: str, params: Any, expected: str
):
    """Ensure that parameter checks work as expected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to check

    params
       The params to check

    expected
       The expected warning message
    """

    with pytest.warns(checks.LspSpecificationWarning, match=expected):
        checks.check_params_against_client_capabilities(capabilities, method, params)


@pytest.mark.parametrize(
    "capabilities,method,result,expected",
    [
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [types.CompletionItem(label="item", commit_characters=["."])],
            "does not support commit characters",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    documentation=types.MarkupContent(
                        value="", kind=types.MarkupKind.Markdown
                    ),
                )
            ],
            "does not support documentation format 'markdown'",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    completion=types.CompletionClientCapabilities()
                )
            ),
            types.TEXT_DOCUMENT_COMPLETION,
            [
                types.CompletionItem(
                    label="item",
                    insert_text_format=types.InsertTextFormat.Snippet,
                )
            ],
            "does not support snippets",
        ),
        (
            types.ClientCapabilities(
                text_document=types.TextDocumentClientCapabilities(
                    document_link=types.DocumentLinkClientCapabilities(
                        tooltip_support=False
                    )
                )
            ),
            types.TEXT_DOCUMENT_DOCUMENT_LINK,
            [types.DocumentLink(range=a_range, tooltip="a tooltip")],
            "does not support tooltips",
        ),
    ],
)
def test_result_check_warning(
    capabilities: types.ClientCapabilities, method: str, result: Any, expected: str
):
    """Ensure that parameter checks work as expected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to check

    params
       The params to check

    expected
       The expected warning message
    """

    with pytest.warns(checks.LspSpecificationWarning, match=expected):
        checks.check_result_against_client_capabilities(capabilities, method, result)



"""Checks to ensure that the server is compliant with the spec.

These can be used in your unit tests to ensure that objects (such as ``CompletionItem``
etc) take into account the client's capabilities when constructed.

However, they are also called automatically during end-to-end tests that make use of the
standard :class:`~pytest_lsp.LanguageClient`. See :ref:`pytest-lsp-spec-checks` for more
details.

"""
import logging
import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

from lsprotocol import types
from pygls.capabilities import get_capability

logger = logging.getLogger(__name__)
ParamsChecker = Callable[[types.ClientCapabilities, Any], None]
ResultChecker = Callable[[types.ClientCapabilities, Any], None]

PARAMS_CHECKS: Dict[str, ParamsChecker] = {}
RESULT_CHECKS: Dict[str, ResultChecker] = {}


class LspSpecificationWarning(UserWarning):
    """Warning raised when encountering results that fall outside the spec."""


def check_result_for(*, method: str) -> Callable[[ResultChecker], ResultChecker]:
    """Define a result check."""

    def defcheck(fn: ResultChecker):
        if (existing := RESULT_CHECKS.get(method, None)) is not None:
            raise ValueError(f"{fn!r} conflicts with existing check {existing!r}")

        RESULT_CHECKS[method] = fn
        return fn

    return defcheck


def check_params_of(*, method: str) -> Callable[[ParamsChecker], ParamsChecker]:
    """Define a params check."""

    def defcheck(fn: ParamsChecker):
        if (existing := PARAMS_CHECKS.get(method, None)) is not None:
            raise ValueError(f"{fn!r} conflicts with existing check {existing!r}")

        PARAMS_CHECKS[method] = fn
        return fn

    return defcheck


def check_result_against_client_capabilities(
    capabilities: Optional[types.ClientCapabilities], method: str, result: Any
):
    """Check that the given result respects the client's declared capabilities.

    This will emit an ``LspSpecificationWarning`` if any issues are detected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to validate the result of

    result
       The result to validate
    """

    if capabilities is None:
        raise RuntimeError("Client has not been initialized")

    # Only run checks if the user provided some capabilities for the client.
    if capabilities == types.ClientCapabilities():
        return

    result_checker = RESULT_CHECKS.get(method, None)
    if result_checker is None:
        return

    try:
        result_checker(capabilities, result)
    except AssertionError as e:
        warnings.warn(str(e), LspSpecificationWarning, stacklevel=4)


def check_params_against_client_capabilities(
    capabilities: Optional[types.ClientCapabilities], method: str, params: Any
):
    """Check that the given params respect the client's declared capabilities.

    This will emit an ``LspSpecificationWarning`` if any issues are detected.

    Parameters
    ----------
    capabilities
       The client's capabilities

    method
       The method name to validate the result of

    params
       The params to validate
    """
    if capabilities is None:
        raise RuntimeError("Client has not been initialized")

    # Only run checks if the user provided some capabilities for the client.
    if capabilities == types.ClientCapabilities():
        return

    params_checker = PARAMS_CHECKS.get(method, None)
    if params_checker is None:
        return

    try:
        params_checker(capabilities, params)
    except AssertionError as e:
        warnings.warn(str(e), LspSpecificationWarning, stacklevel=2)


def check_completion_item(
    item: types.CompletionItem,
    commit_characters_support: bool,
    documentation_formats: Set[str],
    snippet_support: bool,
):
    """Ensure that the given ``CompletionItem`` complies with the given capabilities."""

    if item.commit_characters:
        assert commit_characters_support, "Client does not support commit characters"

    if isinstance(item.documentation, types.MarkupContent):
        kind = item.documentation.kind
        message = f"Client does not support documentation format {kind.value!r}"
        assert kind in documentation_formats, message

    if item.insert_text_format == types.InsertTextFormat.Snippet:
        assert snippet_support, "Client does not support snippets."


@check_result_for(method=types.TEXT_DOCUMENT_COMPLETION)
def completion_items(
    capabilities: types.ClientCapabilities,
    result: Union[types.CompletionList, List[types.CompletionItem], None],
):
    """Ensure that the completion items returned from the server are compliant with the
    spec and the client's declared capabilities."""

    if result is None:
        return

    commit_characters_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.commit_characters_support",
        False,
    )
    documentation_formats = set(
        get_capability(
            capabilities,
            "text_document.completion.completion_item.documentation_format",
            [],
        )
    )
    snippet_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.snippet_support",
        False,
    )

    if isinstance(result, types.CompletionList):
        items = result.items
    else:
        items = result

    for item in items:
        check_completion_item(
            item,
            commit_characters_support=commit_characters_support,
            documentation_formats=documentation_formats,
            snippet_support=snippet_support,
        )


@check_result_for(method=types.COMPLETION_ITEM_RESOLVE)
def completion_item_resolve(
    capabilities: types.ClientCapabilities, item: types.CompletionItem
):
    """Ensure that the completion item returned from the server is compliant with the
    spec and the client's declared capbabilities."""

    commit_characters_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.commit_characters_support",
        False,
    )
    documentation_formats = set(
        get_capability(
            capabilities,
            "text_document.completion.completion_item.documentation_format",
            [],
        )
    )
    snippet_support = get_capability(
        capabilities,
        "text_document.completion.completion_item.snippet_support",
        False,
    )

    check_completion_item(
        item,
        commit_characters_support=commit_characters_support,
        documentation_formats=documentation_formats,
        snippet_support=snippet_support,
    )


@check_result_for(method=types.TEXT_DOCUMENT_DOCUMENT_LINK)
def document_links(
    capabilities: types.ClientCapabilities, result: Optional[List[types.DocumentLink]]
):
    """Ensure that the document links returned from the server are compliant with the
    Spec and the client's declared capabilities."""

    if result is None:
        return

    tooltip_support = get_capability(
        capabilities, "text_document.document_link.tooltip_support", False
    )

    for item in result:
        if item.tooltip:
            assert tooltip_support, "Client does not support tooltips."


@check_params_of(method=types.WINDOW_WORK_DONE_PROGRESS_CREATE)
def work_done_progress_create(
    capabilities: types.ClientCapabilities,
    params: types.WorkDoneProgressCreateParams,
):
    """Assert that the client has support for ``window/workDoneProgress/create``
    requests."""
    is_supported = get_capability(capabilities, "window.workDoneProgress", False)
    assert is_supported, "Client does not support 'window/workDoneProgress/create'"


@check_params_of(method=types.WORKSPACE_CONFIGURATION)
def workspace_configuration(
    capabilities: types.ClientCapabilities,
    params: types.WorkspaceConfigurationParams,
):
    """Ensure that the client has support for ``workspace/configuration`` requests."""
    is_supported = get_capability(capabilities, "workspace.configuration", False)
    assert is_supported, "Client does not support 'workspace/configuration'"



from pygls.protocol import JsonRPCProtocol, default_converter
from pygls.server import Server

server = Server(protocol_cls=JsonRPCProtocol, converter_factory=default_converter)


@server.lsp.fm.feature("math/add")
def addition(ls: Server, params):
    a = params.a
    b = params.b

    ls.lsp.notify("log/message", dict(message=f"{a=}"))
    ls.lsp.notify("log/message", dict(message=f"{b=}"))

    return dict(total=a + b)


@server.lsp.fm.feature("math/sub")
def subtraction(ls: Server, params):
    a = params.a
    b = params.b

    ls.lsp.notify("log/message", dict(message=f"{a=}"))
    ls.lsp.notify("log/message", dict(message=f"{b=}"))

    return dict(total=b - a)


if __name__ == "__main__":
    server.start_io()



import asyncio
import json
import logging
import os
import pathlib
import sys
import traceback
import typing
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from lsprotocol import types
from lsprotocol.converters import get_converter
from packaging.version import parse as parse_version
from pygls.exceptions import JsonRpcException
from pygls.exceptions import PyglsError
from pygls.lsp.client import BaseLanguageClient
from pygls.protocol import default_converter

from .checks import LspSpecificationWarning
from .protocol import LanguageClientProtocol

if sys.version_info.minor < 9:
    import importlib_resources as resources
else:
    import importlib.resources as resources  # type: ignore[no-redef]


__version__ = "0.4.0"
logger = logging.getLogger(__name__)


class LanguageClient(BaseLanguageClient):
    """Used to drive language servers under test."""

    protocol: LanguageClientProtocol

    def __init__(self, *args, configuration: Optional[Dict[str, Any]] = None, **kwargs):
        if "protocol_cls" not in kwargs:
            kwargs["protocol_cls"] = LanguageClientProtocol

        super().__init__("pytest-lsp-client", __version__, *args, **kwargs)

        self.capabilities: Optional[types.ClientCapabilities] = None
        """The client's capabilities."""

        self.shown_documents: List[types.ShowDocumentParams] = []
        """Holds any received show document requests."""

        self.messages: List[types.ShowMessageParams] = []
        """Holds any received ``window/showMessage`` requests."""

        self.log_messages: List[types.LogMessageParams] = []
        """Holds any received ``window/logMessage`` requests."""

        self.diagnostics: Dict[str, List[types.Diagnostic]] = {}
        """Holds any recieved diagnostics."""

        self.progress_reports: Dict[
            types.ProgressToken, List[types.ProgressParams]
        ] = {}
        """Holds any received progress updates."""

        self.error: Optional[Exception] = None
        """Indicates if the client encountered an error."""

        config = (configuration or {"": {}}).copy()
        if "" not in config:
            config[""] = {}

        self._configuration: Dict[str, Dict[str, Any]] = config
        """Holds ``workspace/configuration`` values."""

        self._setup_log_index = 0
        """Used to keep track of which log messages occurred during startup."""

        self._last_log_index = 0
        """Used to keep track of which log messages correspond with which test case."""

    async def server_exit(self, server: asyncio.subprocess.Process):
        """Called when the server process exits."""
        logger.debug("Server process exited with code: %s", server.returncode)

        if self._stop_event.is_set():
            return

        stderr = ""
        if server.stderr is not None:
            stderr_bytes = await server.stderr.read()
            stderr = stderr_bytes.decode("utf8")

        loop = asyncio.get_running_loop()
        loop.call_soon(
            cancel_all_tasks,
            f"Server process exited with return code: {server.returncode}\n{stderr}",
        )

    def report_server_error(
        self, error: Exception, source: Union[PyglsError, JsonRpcException]
    ):
        """Called when the server does something unexpected, e.g. sending malformed
        JSON."""
        self.error = error
        tb = "".join(traceback.format_exc())

        message = f"{source.__name__}: {error}\n{tb}"  # type: ignore

        loop = asyncio.get_running_loop()
        loop.call_soon(cancel_all_tasks, message)

        if self._stop_event:
            self._stop_event.set()

    def get_configuration(
        self, *, section: Optional[str] = None, scope_uri: Optional[str] = None
    ) -> Optional[Any]:
        """Get a configuration value.

        Parameters
        ----------
        section
           The optional section name to retrieve.
           If ``None`` the top level configuration object for the requested scope will
           be returned

        scope_uri
           The scope at which to set the configuration.
           If ``None``, this will default to the global scope.

        Returns
        -------
        Optional[Any]
           The requested configuration value or ``None`` if not found.
        """
        section = section or ""
        scope = scope_uri or ""

        # Find the longest prefix of ``scope``. The empty string is a prefix of all
        # strings so there will always be at least one match
        candidates = [c for c in self._configuration.keys() if scope.startswith(c)]
        selected = sorted(candidates, key=len, reverse=True)[0]

        if (item := self._configuration.get(selected, None)) is None:
            return None

        if section == "":
            return item

        for segment in section.split("."):
            if not hasattr(item, "get"):
                return None

            if (item := item.get(segment, None)) is None:
                return None

        return item

    def set_configuration(
        self,
        item: Any,
        *,
        section: Optional[str] = None,
        scope_uri: Optional[str] = None,
    ):
        """Set a configuration value.

        Parameters
        ----------
        item
           The value to set

        section
           The optional section name to set.
           If ``None`` the top level configuration object will be overriden with
           ``item``.

        scope_uri
           The scope at which to set the configuration.
           If ``None``, this will default to the global scope.
        """
        section = section or ""
        scope = scope_uri or ""

        if section == "":
            self._configuration[scope] = item
            return

        config = self._configuration.setdefault(scope, {})
        *parents, name = section.split(".")

        for segment in parents:
            config = config.setdefault(segment, {})

        config[name] = item

    async def initialize_session(
        self, params: types.InitializeParams
    ) -> types.InitializeResult:
        """Make an ``initialize`` request to a lanaguage server.

        It will also automatically send an ``initialized`` notification once
        the server responds.

        Parameters
        ----------
        params
           The parameters to send to the client.

           The following fields will be automatically set if left blank.

           - ``process_id``: Set to the PID of the current process.

        Returns
        -------
        InitializeResult
           The result received from the client.
        """
        self.capabilities = params.capabilities

        if params.process_id is None:
            params.process_id = os.getpid()

        response = await self.initialize_async(params)
        self.initialized(types.InitializedParams())

        return response

    async def shutdown_session(self) -> None:
        """Shutdown the server under test.

        Helper method that handles sending ``shutdown`` and ``exit`` messages in the
        correct order.

        .. note::

           This method will not attempt to send these messages if a fatal error has
           occurred.

        """
        if self.error is not None or self.capabilities is None:
            return

        await self.shutdown_async(None)
        self.exit(None)

    async def wait_for_notification(self, method: str):
        """Block until a notification with the given method is received.

        Parameters
        ----------
        method
           The notification method to wait for, e.g. ``textDocument/publishDiagnostics``
        """
        return await self.protocol.wait_for_notification_async(method)


def cancel_all_tasks(message: str):
    """Called to cancel all awaited tasks."""

    for task in asyncio.all_tasks():
        if sys.version_info.minor < 9:
            task.cancel()
        else:
            task.cancel(message)


def make_test_lsp_client() -> LanguageClient:
    """Construct a new test client instance with the handlers needed to capture
    additional responses from the server."""

    client = LanguageClient(
        converter_factory=default_converter,
    )

    @client.feature(types.WORKSPACE_CONFIGURATION)
    def configuration(client: LanguageClient, params: types.ConfigurationParams):
        return [
            client.get_configuration(section=item.section, scope_uri=item.scope_uri)
            for item in params.items
        ]

    @client.feature(types.TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
    def publish_diagnostics(
        client: LanguageClient, params: types.PublishDiagnosticsParams
    ):
        client.diagnostics[params.uri] = params.diagnostics

    @client.feature(types.WINDOW_WORK_DONE_PROGRESS_CREATE)
    def create_work_done_progress(
        client: LanguageClient, params: types.WorkDoneProgressCreateParams
    ):
        if params.token in client.progress_reports:
            # TODO: Send an error reponse to the client - might require changes
            #       to pygls...
            warnings.warn(
                f"Duplicate progress token: {params.token!r}", LspSpecificationWarning
            )

        client.progress_reports.setdefault(params.token, [])
        return None

    @client.feature(types.PROGRESS)
    def progress(client: LanguageClient, params: types.ProgressParams):
        if params.token not in client.progress_reports:
            warnings.warn(
                f"Unknown progress token: {params.token!r}", LspSpecificationWarning
            )

        if not params.value:
            return

        if (kind := params.value.get("kind", None)) == "begin":
            type_: Type[Any] = types.WorkDoneProgressBegin
        elif kind == "report":
            type_ = types.WorkDoneProgressReport
        elif kind == "end":
            type_ = types.WorkDoneProgressEnd
        else:
            raise TypeError(f"Unknown progress kind: {kind!r}")

        value = client.protocol._converter.structure(params.value, type_)
        client.progress_reports.setdefault(params.token, []).append(value)

    @client.feature(types.WINDOW_LOG_MESSAGE)
    def log_message(client: LanguageClient, params: types.LogMessageParams):
        client.log_messages.append(params)

        levels = [logger.error, logger.warning, logger.info, logger.debug]
        levels[params.type.value - 1](params.message)

    @client.feature(types.WINDOW_SHOW_MESSAGE)
    def show_message(client: LanguageClient, params):
        client.messages.append(params)

    @client.feature(types.WINDOW_SHOW_DOCUMENT)
    def show_document(
        client: LanguageClient, params: types.ShowDocumentParams
    ) -> types.ShowDocumentResult:
        client.shown_documents.append(params)
        return types.ShowDocumentResult(success=True)

    return client


def client_capabilities(client_spec: str) -> types.ClientCapabilities:
    """Find the capabilities that correspond to the given client spec.

    This function supports the following syntax

    ``client-name`` or ``client-name@latest``
       Return the capabilities of the latest version of ``client-name``

    ``client-name@v2``
       Return the latest release of the ``v2`` of ``client-name``

    ``client-name@v2.3.1``
       Return exactly ``v2.3.1`` of ``client-name``

    Parameters
    ----------
    client_spec
       The string describing the client to load the corresponding
       capabilities for.

    Raises
    ------
    ValueError
       If the requested client's capabilities could not be found

    Returns
    -------
    ClientCapabilities
       The requested client capabilities
    """

    candidates: Dict[str, pathlib.Path] = {}

    client_spec = client_spec.replace("-", "_")
    target_version = "latest"

    if "@" in client_spec:
        client_spec, target_version = client_spec.split("@")
        if target_version.startswith("v"):
            target_version = target_version[1:]

    for resource in resources.files("pytest_lsp.clients").iterdir():
        filename = typing.cast(pathlib.Path, resource)

        # Skip the README or any other files that we don't care about.
        if not filename.suffix == ".json":
            continue

        name, version = filename.stem.split("_v")
        if name == client_spec:
            if version.startswith(target_version) or target_version == "latest":
                candidates[version] = filename

    if len(candidates) == 0:
        raise ValueError(
            f"Could not find capabilities for '{client_spec}@{target_version}'"
        )

    # Out of the available candidates, choose the latest version
    selected_version = sorted(candidates.keys(), key=parse_version, reverse=True)[0]
    filename = candidates[selected_version]

    converter = get_converter()
    capabilities = json.loads(filename.read_text())

    params = converter.structure(capabilities, types.InitializeParams)
    logger.info(
        "Selected %s v%s",
        params.client_info.name,  # type: ignore[union-attr]
        params.client_info.version,  # type: ignore[union-attr]
    )

    return params.capabilities



from lsprotocol.types import TEXT_DOCUMENT_COMPLETION
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionParams
from pygls.server import LanguageServer

server = LanguageServer("window-log-message", "v1")


@server.feature(TEXT_DOCUMENT_COMPLETION)
def completion(ls: LanguageServer, params: CompletionParams):
    items = []

    for i in range(10):
        ls.show_message_log(f"Suggesting item {i}")
        items.append(CompletionItem(label=f"item-{i}"))

    return items


if __name__ == "__main__":
    server.start_io()



from lsprotocol.types import TEXT_DOCUMENT_DID_OPEN
from lsprotocol.types import Diagnostic
from lsprotocol.types import DidOpenTextDocumentParams
from lsprotocol.types import Position
from lsprotocol.types import Range
from pygls.server import LanguageServer

server = LanguageServer("diagnostic-server", "v1")


@server.feature(TEXT_DOCUMENT_DID_OPEN)
def did_open(ls: LanguageServer, params: DidOpenTextDocumentParams):
    ls.publish_diagnostics(
        params.text_document.uri,
        [
            Diagnostic(
                message="There is an error here.",
                range=Range(
                    start=Position(line=1, character=1),
                    end=Position(line=1, character=10),
                ),
            )
        ],
    )


if __name__ == "__main__":
    server.start_io()
